from fastapi_advanced_filters.filter_metaclass.helpers.field_criteria import (
    attrs_to_field_criteria,
    from_field_criteria_to_attr,
)
from fastapi_advanced_filters.filter_metaclass.helpers.pagination import (
    generate_annotations_for_pagination,
)
from fastapi_advanced_filters.filter_metaclass.helpers.q_search import (
    generate_annotations_for_qsearch,
)
from fastapi_advanced_filters.filter_metaclass.helpers.selectable import (
    generate_annotations_for_selectable_fields,
)
from fastapi_advanced_filters.filter_metaclass.helpers.sortable import (
    generate_annotations_for_sortable_fields,
)



from typing import Annotated, Any, Dict, Optional

from pydantic.fields import Field

from fastapi_advanced_filters.enums import PaginationEnum


def generate_annotations_for_pagination(
    filter_config_cls: type,
) -> Optional[Dict[str, Any]]:
    if (
        not hasattr(filter_config_cls, "pagination")
        or not filter_config_cls.pagination
        or not isinstance(filter_config_cls.pagination, PaginationEnum)
    ):
        return None
    if filter_config_cls.pagination == PaginationEnum.OFFSET_BASED:
        return {
            "limit": Annotated[
                int,
                Field(
                    default=100,
                    alias="limit",
                    title="Limit the number of results",
                    description="Limit the number of results returned. Default is 100.",
                    ge=1,
                    le=1000,
                ),
            ],
            "offset": Annotated[
                int,
                Field(
                    default=0,
                    alias="offset",
                    title="Offset the results",
                    description="Offset the results by a certain number. Default is 0.",
                    ge=0,
                ),
            ],
        }
    return {
        "page": Annotated[
            int,
            Field(
                default=1,
                alias="page",
                title="Page number",
                description="Page number for pagination. Default is 1.",
                ge=1,
            ),
        ],
        "page_size": Annotated[
            int,
            Field(
                default=100,
                alias="size",
                title="Number of items per page",
                description="Number of items per page for pagination. Default is 100.",
                ge=1,
                le=1000,
            ),
        ],
    }



from dataclasses import dataclass
from typing import Any

from annotated_types import BaseMetadata

from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum


@dataclass(frozen=True)
class AdvancedQSearch(BaseMetadata):
    model_attrs_with_op: dict[OperationEnum, list[Any]]
    logical_op: LogicalOperator = LogicalOperator.OR



from fastapi_advanced_filters.data_classes.advanced_qsearch import AdvancedQSearch
from fastapi_advanced_filters.data_classes.field_criteria import FieldCriteria
from fastapi_advanced_filters.data_classes.filter_result import FilterResult
from fastapi_advanced_filters.data_classes.pagination import Pagination
from fastapi_advanced_filters.data_classes.qsearch import QSearch
from fastapi_advanced_filters.data_classes.selectable import Selectable
from fastapi_advanced_filters.data_classes.sortby import SortBy

__all__ = [
    "AdvancedQSearch",
    "FieldCriteria",
    "FilterResult",
    "Pagination",
    "QSearch",
    "Selectable",
    "SortBy",
]



from fastapi_advanced_filters.data_classes import (
    AdvancedQSearch,
    FieldCriteria,
    FilterResult,
    Pagination,
    QSearch,
    Selectable,
    SortBy,
)
from fastapi_advanced_filters.enums import (
    LogicalOperator,
    OperationEnum,
    OrderEnum,
    PaginationEnum,
)
from fastapi_advanced_filters.filters import BaseFilter
from fastapi_advanced_filters.operation_mapping import (
    SQLALCHEMY_LOGICAL_OP_MAPPING,
    SQLALCHEMY_OP_MAPPING,
    SQLALCHEMY_SORTING_MAPPING,
)

__all__ = [
    "BaseFilter",
    "OperationEnum",
    "FieldCriteria",
    "Selectable",
    "SortBy",
    "QSearch",
    "PaginationEnum",
    "OrderEnum",
    "FilterResult",
    "Pagination",
    "LogicalOperator",
    "AdvancedQSearch",
    "SQLALCHEMY_OP_MAPPING",
    "SQLALCHEMY_SORTING_MAPPING",
    "SQLALCHEMY_LOGICAL_OP_MAPPING",
]



from typing import Any, Callable

from pydantic.fields import FieldInfo

from fastapi_advanced_filters.data_classes import AdvancedQSearch, QSearch
from fastapi_advanced_filters.enums import OperationEnum


class QSearchMixin:
    __op_mapping__: dict[OperationEnum, Callable[..., Any]]
    __logical_op_mapping__: dict[Any, Callable[..., Any]]

    def __get_q_search_metadata(self, attr_name: str) -> QSearch | AdvancedQSearch:
        # Access model_fields on the class to avoid Pydantic V2.11 deprecation warnings
        field: FieldInfo = type(self).model_fields.get(attr_name)  # type: ignore
        assert (
            hasattr(field, "metadata")
        ) and field.metadata, f"Field '{attr_name}' has no metadata."
        assert isinstance(field.metadata[0], (QSearch, AdvancedQSearch)), (
            f"Field '{attr_name}' metadata is not of type "
            f"(QSearch, AdvancedQSearch)."
        )
        return field.metadata[0]

    def build_q_search(self, attr_name: str = "q_search") -> Any | None:
        if not hasattr(self, attr_name) or not getattr(self, attr_name):
            return None
        field_metadata = self.__get_q_search_metadata(attr_name)
        q_search: str = getattr(self, attr_name)
        assert field_metadata is not None and isinstance(
            field_metadata, (QSearch, AdvancedQSearch)
        ), (
            f"Field '{attr_name}' must have metadata of type "
            f"(QSearch, AdvancedQSearch)."
        )
        if isinstance(field_metadata, QSearch):
            return self.__build_q_search_operation(field_metadata, q_search)
        return self.__build_advanced_q_search_operation(field_metadata, q_search)

    def __build_q_search_operation(self, field: QSearch, value: str) -> Any:
        condition: Callable[..., Any] | None = self.__op_mapping__.get(field.op, None)
        assert (
            field.model_attrs is not None and condition is not None
        ), f"Field '{field.name}' must have 'model_attrs' and valid 'op' defined."
        conditions = [
            condition(model_attr, value)
            for model_attr in field.model_attrs
            if model_attr is not None
        ]
        return (
            self.__logical_op_mapping__[field.logical_op](*conditions)
            if conditions
            else None
        )

    def __build_advanced_q_search_operation(
        self, field: AdvancedQSearch, value: str
    ) -> Any:
        assert (
            field.model_attrs_with_op is not None
        ), f"Field '{field.name}' must have 'model_attrs_with_op' defined."
        assert (
            field.logical_op is not None
        ), f"Field '{field.name}' must have 'logical_op' defined."
        conditions = []
        for op, model_attrs in field.model_attrs_with_op.items():
            if (condition := self.__op_mapping__.get(op, None)) is not None:
                conditions.extend(
                    [
                        condition(model_attr, value)
                        for model_attr in model_attrs
                        if model_attr is not None
                    ]
                )

        return (
            self.__logical_op_mapping__[field.logical_op](*conditions)
            if conditions
            else None
        )



import operator
from datetime import date, datetime
from typing import Any, Callable

try:
    from sqlalchemy import ARRAY, Boolean, Column, String, and_, or_
except ImportError:  # pragma: no cover
    ARRAY = Boolean = Column = or_ = and_ = String = None  # type: ignore

from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum, OrderEnum


def in_funct(field: Any, values: str) -> Any:  # type: ignore
    try:
        conditions = []
        if ARRAY is not None and isinstance(getattr(field, "type", None), ARRAY):
            return field.contains([val for val in values.split(",") if val])
        for value in values.split(","):
            if Boolean is not None and isinstance(field.type, Boolean):
                conditions.append(False if value == "false" or value == "0" else True)
            else:
                conditions.append(field.type.python_type(value))
        return field.in_(conditions)
    except ValueError:
        return None


def not_in_funct(field: Any, values: str) -> Any:  # type: ignore
    try:
        conditions = []
        for value in values.split(","):
            if Boolean is not None and isinstance(field.type, Boolean):
                conditions.append(False if value == "false" or value == "0" else True)
            else:
                conditions.append(field.type.python_type(value))
        return ~field.in_(conditions)
    except ValueError:
        return None


def contains(field: Any, val: str) -> Any:  # type: ignore
    if ARRAY is not None and isinstance(getattr(field, "type", None), ARRAY):
        return field.any([val for val in val.split(",") if val])
    if String is not None and isinstance(field.type, String):
        return or_(field.contains(val) for val in val.split(",") if val)
    return or_(field == i for i in val.split(",") if i)


def between(field: Any, values: str) -> Any:  # type: ignore
    parts = [p for p in values.split(",") if p]
    if len(parts) >= 2 and getattr(field.type, "python_type", None) in (int, float):
        split_values = list(map(int, parts))
        return field.between(min(split_values), max(split_values))
    if len(parts) >= 2 and getattr(field.type, "python_type", None) in (
        date,
        datetime,
    ):
        # parse into date if date-like, otherwise datetime.fromisoformat
        # would be fallback
        try:
            split_dates = list(map(date.fromisoformat, parts))
        except ValueError:
            split_dates = list(map(lambda x: datetime.fromisoformat(x).date(), parts))
        return field.between(min(split_dates), max(split_dates))
    raise ValueError("Invalid values for between operation")


OP_MAPPING: dict[OperationEnum, Callable[..., Any]] = {
    OperationEnum.LIKE: lambda x, y: x.like(f"%{y}%"),
    OperationEnum.ILIKE: lambda x, y: x.ilike(f"%{y}%"),
    OperationEnum.EQ: operator.eq,
    OperationEnum.NEQ: operator.ne,
    OperationEnum.GT: operator.gt,
    OperationEnum.GTE: operator.ge,
    OperationEnum.LTE: operator.le,
    OperationEnum.LT: operator.lt,
    OperationEnum.IS: lambda x, y: x.is_(y),
    OperationEnum.IN: in_funct,
    OperationEnum.CONT: contains,
    OperationEnum.NOTIN: not_in_funct,
    OperationEnum.ISNULL: lambda x, y: x.is_(None) if y else x.is_not(None),
    OperationEnum.BTW: between,
}

SORTING_MAPPING: dict[OrderEnum, Callable[[Any], Any]] = {
    OrderEnum.ASC: lambda x: x.asc(),
    OrderEnum.DESC: lambda x: x.desc(),
}

LOGICAL_OP_MAPPING: dict[LogicalOperator, Callable[..., Any]] = {
    LogicalOperator.AND: and_,
    LogicalOperator.OR: or_,
}



from typing import Any

from pydantic.fields import FieldInfo

from fastapi_advanced_filters.data_classes import Selectable


class SelectMixin:
    def __get_selectable_metadata(self, attr_name: str) -> Selectable:
        # Access model_fields on the class to avoid Pydantic V2.11 deprecation warnings
        field: FieldInfo = type(self).model_fields.get(attr_name)  # type: ignore
        assert (
            hasattr(field, "metadata")
        ) and field.metadata, f"Field '{attr_name}' has no metadata."
        assert isinstance(
            field.metadata[0], Selectable
        ), f"Field '{attr_name}' metadata is not of type {Selectable.__name__}."
        return field.metadata[0]

    def build_selectable_fields(self, attr_name: str = "select") -> list[Any] | None:
        if not hasattr(self, attr_name) or not getattr(self, attr_name):
            return None
        field_metadata: Selectable = self.__get_selectable_metadata(attr_name)
        select: str | list[str] = getattr(self, attr_name)
        if select == "all":
            return list(field_metadata.model_attrs.values())
        fields: list[Any] = []
        for field in select:
            if field_attr := field_metadata.get_attr(field):
                fields.append(field_attr)
        return fields if fields else None



from typing import Annotated, Any, Dict, Generator, TypeVar, Union

from pydantic.fields import Field, FieldInfo

try:
    from sqlalchemy.orm.attributes import InstrumentedAttribute
except ImportError:  # pragma: no cover
    InstrumentedAttribute = TypeVar("InstrumentedAttribute")  # type: ignore

from fastapi_advanced_filters.data_classes import FieldCriteria
from fastapi_advanced_filters.enums import OperationEnum


def attrs_to_field_criteria(
    model_cls: type,
    fields: list[str] | str | None = None,
    prefix: str | None = None,
    op: tuple[OperationEnum, ...] | None = None,
) -> Generator[FieldCriteria, None, None]:
    # Normalize fields into a list of attribute names
    if fields is None:
        model_attrs: list[Any] = []
    elif fields == "__all__":
        model_attrs = __get_all_fields_from_model(model_cls)
    elif isinstance(fields, str):
        model_attrs = [fields]
    else:
        model_attrs = fields

    for attr in model_attrs:
        if isinstance(attr, FieldCriteria):
            yield attr
            continue

        field: Any = getattr(model_cls, attr)

        yield __get_field_criteria_from_model(
            field_type=__type_from_attr(field),
            model_attr=field,
            field_name=attr,
            prefix=prefix,
            op=op or (OperationEnum.EQ,),
        )


def __type_from_attr(attr: Any) -> type:
    if hasattr(attr, "property") and hasattr(attr.property, "columns"):
        column = attr.property.columns[0]
        return column.type.python_type
    return str  # Default to str if type cannot be determined


def __get_all_fields_from_model(model_cls: type) -> list[str]:
    assert model_cls is not None, "Model class cannot be None."
    return [
        attr
        for attr in dir(model_cls)
        if not attr.startswith("_")
        and (
            isinstance(getattr(model_cls, attr), FieldInfo)
            or isinstance(getattr(model_cls, attr), InstrumentedAttribute)
        )
    ]


def __get_field_criteria_from_model(
    field_type: type,
    model_attr: type,
    field_name: str,
    prefix: str | None,
    op: tuple[OperationEnum, ...],
) -> FieldCriteria:
    return FieldCriteria(
        name=field_name,
        field_type=field_type,
        model_attr=model_attr,
        op=op,
        prefix=prefix,
    )


def from_field_criteria_to_attr(
    field_criteria: FieldCriteria,
) -> Dict[str, Any]:
    fields: Dict[str, Any] = {}
    for op in field_criteria.op:
        annotation_type: type = field_criteria.field_type
        field_name: str = field_criteria.get_field_name(op)
        is_required: bool = (
            field_criteria.required_op is not None and op in field_criteria.required_op
        )
        kwargs: dict[str, Any] = field_criteria.fields_kwargs.copy()
        if is_required:
            kwargs["default"] = ...  # type: ignore
        else:
            kwargs["default"] = None
        fields[field_name] = Annotated[
            Union[annotation_type, str, None],
            Field(
                alias=(field_criteria.get_alias_name(op)),
                title=f"Filter by {field_criteria.name} with operation {op}",
                description=f"Filter by {field_criteria.name} with operation {op}",
                **kwargs,
            ),
            op,
            field_criteria,
        ]
    return fields



from .filter import FilterMixin
from .pagination import PaginationMixin
from .q_search import QSearchMixin
from .selecting import SelectMixin
from .sorting import SortingMixin

__all__ = [
    "FilterMixin",
    "SortingMixin",
    "SelectMixin",
    "QSearchMixin",
    "PaginationMixin",
]



from typing import Callable, Optional

from fastapi_advanced_filters.enums import OrderEnum


def to_camel_case(snake_str: str) -> str:
    if not snake_str:
        return ""
    components = snake_str.split("_")
    first = components[0].lower() if not components[0].isupper() else components[0]
    return first + "".join(x.title() for x in components[1:])


def to_snake_case(camel_str: str) -> str:
    if not camel_str:
        return ""
    snake_str = ""
    prev_char_is_upper = False
    for i, char in enumerate(camel_str):
        if char.isupper():
            if snake_str and not prev_char_is_upper and i > 0:
                snake_str += "_"
            snake_str += char.lower()
            prev_char_is_upper = True
        else:
            snake_str += char
            prev_char_is_upper = False
    return snake_str


def extract_sorting(sort_field: str) -> tuple[str, OrderEnum]:
    sort_field_with_order_dir: str = sort_field.strip()
    if sort_field_with_order_dir.startswith("-"):
        return sort_field_with_order_dir[1:], OrderEnum.DESC
    return sort_field_with_order_dir, OrderEnum.ASC


def validate_sortable_schema(
    allowed_sortable_attr: list[str],
) -> Callable[[Optional[str]], list[tuple[str, OrderEnum]]]:
    def validate_schema(
        provided_sortable_attr: Optional[str],
    ) -> list[tuple[str, OrderEnum]]:
        if provided_sortable_attr is None:
            return []
        formatted_sortable_data: list[tuple[str, OrderEnum]] = []
        for sort_field_with_order_dir in provided_sortable_attr.strip().split(","):
            if not sort_field_with_order_dir:
                raise ValueError("Invalid sorting format")
            sort_field, op = extract_sorting(sort_field_with_order_dir)
            if sort_field not in allowed_sortable_attr:
                raise ValueError(f"Field '{sort_field}' is not sortable")
            formatted_sortable_data.append((sort_field, op))
        return formatted_sortable_data

    return validate_schema


def validate_selectable_schema(
    allowed_selectable_attr: list[str],
) -> Callable[[str], list[str]]:
    def validate_schema(provided_selecteble_attr: str) -> list[str]:
        if provided_selecteble_attr == "all":
            return allowed_selectable_attr
        formatted_selectable_data: list[str] = []
        for selected_field in provided_selecteble_attr.split(","):
            if not selected_field:
                raise ValueError(
                    "select must be 'all', a list of strings, or a "
                    "comma-separated string"
                )
            if selected_field not in allowed_selectable_attr:
                raise ValueError(f"Field '{selected_field}' is not selectable")
            formatted_selectable_data.append(selected_field)
        return formatted_selectable_data

    return validate_schema



from dataclasses import dataclass, field
from typing import Any

from annotated_types import BaseMetadata

from fastapi_advanced_filters.utils import to_camel_case


@dataclass(frozen=True)
class Selectable(BaseMetadata):
    model_attrs: dict[str, Any]
    __camelcase_alias_cache: dict[str, Any] = field(default_factory=dict, init=False)
    alias_as_camelcase: bool = False

    def __post_init__(self):
        if self.model_attrs and self.alias_as_camelcase:
            object.__setattr__(
                self,
                "_Selectable__camelcase_alias_cache",
                {to_camel_case(name): attr for name, attr in self.model_attrs.items()},
            )

    def get_names(self) -> list[str]:
        if self.alias_as_camelcase:
            return list(self.__camelcase_alias_cache.keys())
        return list(self.model_attrs.keys())

    def get_attr(self, sort_attr: str) -> Any:
        if self.alias_as_camelcase:
            return self.__camelcase_alias_cache.get(sort_attr)
        return self.model_attrs.get(sort_attr)



from typing import Annotated, Any, Dict, Optional

from pydantic import AfterValidator, Field

from fastapi_advanced_filters.data_classes import Selectable
from fastapi_advanced_filters.utils import validate_selectable_schema


def generate_annotations_for_selectable_fields(
    filter_config_cls: type | Selectable,
) -> Optional[Dict[str, Any]]:
    if (
        hasattr(filter_config_cls, "select_only")
        and isinstance(filter_config_cls.select_only, Selectable)
        and (selectable := filter_config_cls.select_only)
    ):
        return {
            "select": Annotated[
                str,
                Field(
                    default="all",
                    alias="select",
                    title=(
                        "Select specific fields ("
                        f"{', '.join(selectable.get_names())} or 'all')"
                    ),
                    description=(
                        "Comma separated fields to include in the response or 'all'"
                        " to include all fields."
                    ),
                ),
                selectable,
                AfterValidator(validate_selectable_schema(selectable.get_names())),
            ]
        }
    return None



from typing import Any

from pydantic import BaseModel

from fastapi_advanced_filters.enums import OperationEnum
from fastapi_advanced_filters.filter_metaclass.helpers import (
    attrs_to_field_criteria,
    from_field_criteria_to_attr,
    generate_annotations_for_pagination,
    generate_annotations_for_qsearch,
    generate_annotations_for_selectable_fields,
    generate_annotations_for_sortable_fields,
)


class FilterMetaClass(type(BaseModel)):  # type: ignore
    """Metaclass to dynamically create filter fields from configuration.

    This processes a nested `FilterConfig` inside a Pydantic model and generates
    filter fields with appropriate types, aliases, and metadata for filtering
    operations.
    """

    def __new__(
        mcs,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, Any],
        **kwargs: dict[str, Any],
    ) -> type:
        if attrs and "FilterConfig" in attrs:
            annotations: dict[str, Any] | None = attrs.get("__annotations__", None)
            if annotations is None:
                attrs["__annotations__"] = mcs.generate_annotations_for_filters(
                    attrs["FilterConfig"]
                )
            else:
                attrs["__annotations__"].update(
                    mcs.generate_annotations_for_filters(attrs["FilterConfig"])
                )
        return super().__new__(mcs, name, bases, attrs, **kwargs)

    def generate_annotations_for_filters(  # noqa: C901
        filter_config_cls: type,
    ) -> dict[str, Any]:
        annotations: dict[str, Any] = {}
        if hasattr(filter_config_cls, "fields"):
            fields = filter_config_cls.fields
            if hasattr(filter_config_cls, "model"):
                fields = attrs_to_field_criteria(
                    model_cls=filter_config_cls.model,
                    fields=filter_config_cls.fields,
                    prefix=getattr(filter_config_cls, "prefix", None),
                    op=getattr(filter_config_cls, "default_op", (OperationEnum.EQ,)),
                )

            for field in fields:
                annotations.update(from_field_criteria_to_attr(field))
        if q_search_annotation := generate_annotations_for_qsearch(filter_config_cls):
            annotations.update(q_search_annotation)
        if selectable_annotation := generate_annotations_for_selectable_fields(
            filter_config_cls
        ):
            annotations.update(selectable_annotation)
        if sortable_annotation := generate_annotations_for_sortable_fields(
            filter_config_cls
        ):
            annotations.update(sortable_annotation)
        if pagination_annotations := generate_annotations_for_pagination(
            filter_config_cls
        ):
            annotations.update(pagination_annotations)
        return annotations



from typing import Any

from pydantic.fields import FieldInfo

from fastapi_advanced_filters.data_classes import SortBy
from fastapi_advanced_filters.enums import OrderEnum


class SortingMixin:
    __sorting_mapping__: dict[OrderEnum, Any]

    def __get_sorting_metadata(self, attr_name: str) -> SortBy:
        # Access model_fields on the class to avoid Pydantic V2.11 deprecation warnings
        field: FieldInfo = type(self).model_fields.get(attr_name)  # type: ignore
        assert (
            hasattr(field, "metadata")
        ) and field.metadata, f"Field '{attr_name}' has no metadata."
        assert isinstance(
            field.metadata[0], SortBy
        ), f"Field '{attr_name}' metadata is not of type {SortBy.__name__}."
        return field.metadata[0]

    def build_sorting(self, attr_name: str = "sorting") -> list[Any] | None:
        if not hasattr(self, attr_name) or not getattr(self, attr_name):
            return None
        field_metadata: SortBy = self.__get_sorting_metadata(attr_name)
        sorting: list[tuple[str, OrderEnum]] = getattr(self, attr_name)
        sort_by: list[Any] = []
        for sort_field, op in sorting:
            field_attr = field_metadata.get_attr(sort_field)
            if field_attr and self.__sorting_mapping__.get(op) is not None:
                sort_by.append(self.__sorting_mapping__[op](field_attr))
        return sort_by if sort_by else None



from enum import StrEnum


class PaginationEnum(StrEnum):
    PAGE_BASED = "page_based"
    OFFSET_BASED = "offset_based"


class LogicalOperator(StrEnum):
    AND = "AND"
    OR = "OR"


class OperationEnum(StrEnum):
    ILIKE = "ilike"
    LIKE = "like"
    EQ = "eq"
    NEQ = "neq"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    IN = "in"
    NOTIN = "notin"
    ISNULL = "isnull"
    BTW = "btw"
    CONT = "cont"
    IS = "is"


class OrderEnum(StrEnum):
    ASC = "asc"
    DESC = "desc"



from datetime import datetime
from typing import Any

from pydantic.fields import FieldInfo

from fastapi_advanced_filters.data_classes import FieldCriteria
from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum


class FilterMixin:
    __op_mapping__: dict[OperationEnum, Any]
    __logical_op_mapping__: dict[LogicalOperator, Any]

    def __get_filterable_field_metadata(
        self, attr_name: str
    ) -> tuple[OperationEnum, FieldCriteria] | None:
        # Access model_fields on the class to avoid Pydantic V2.11 deprecation warnings
        field: FieldInfo = type(self).model_fields.get(attr_name)  # type: ignore
        if (
            hasattr(field, "metadata")
            and field.metadata
            and isinstance(field.metadata, list)
            and len(field.metadata) == 2
        ) and (
            (operation := field.metadata[0])
            and (field_metadata := field.metadata[1])
            and isinstance(field_metadata, FieldCriteria)
        ):
            return operation, field_metadata
        return None

    def __build_operation(
        self, operation: OperationEnum, field: FieldCriteria, value: Any
    ) -> list[Any] | Any:
        assert (
            field.model_attr is not None
            or field.custom_filter_per_op is not None
            or field.model_attrs_with_logical_op is not None
        ), (
            f"Field '{field.name}' must have either 'model_attr', "
            f"'custom_filter_per_op' or 'model_attrs_with_logical_op' defined."
        )
        if isinstance(value, datetime):
            value = value.replace(tzinfo=None)
        if field.custom_filter_per_op and callable(field.custom_filter_per_op):
            return field.custom_filter_per_op(operation, value)
        if (
            self.__op_mapping__.get(operation, None) is not None
            and field.model_attr is not None
        ):
            assert isinstance(field.model_attr, (list, tuple)) is False, (
                f"Field '{field.name}' has multiple 'model_attr' defined. "
                f"Use 'model_attrs_with_logical_op' instead."
            )
            condition = self.__op_mapping__[operation](field.model_attr, value)
            return condition
        if (
            field.model_attrs_with_logical_op is not None
            and self.__op_mapping__.get(operation, None) is not None
        ):
            assert isinstance(field.model_attrs_with_logical_op, (tuple, list)), (
                f"Field '{field.name}' must have 'model_attrs_with_logical_op' "
                f"defined as a list."
            )
            model_attrs, logical_operator = field.model_attrs_with_logical_op
            assert isinstance(model_attrs, (list, tuple)), (
                f"Field '{field.name}' must have list of 'model_attrs' defined in "
                f"'model_attrs_with_logical_op'."
            )
            assert (
                isinstance(logical_operator, LogicalOperator)
                and self.__logical_op_mapping__.get(logical_operator, None) is not None
            ), (
                f"Field '{field.name}' must have 'logical_operator' defined in "
                f"'model_attrs_with_logical_op'."
            )
            conditions: list[Any] = []
            for model_attr in model_attrs:
                if (condition := self.__op_mapping__.get(operation, None)) is not None:
                    conditions.append(condition(model_attr, value))
            return (
                self.__logical_op_mapping__[logical_operator](*conditions)
                if conditions
                else None
            )
        return None

    def build_filters(self) -> list[Any] | None:
        filters: list[Any] = []
        for key, value in self.model_dump(  # type: ignore
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
            exclude=(
                "sorting",
                "select",
                "q_search",
                "limit",
                "offset",
                "page",
                "page_size",
            ),
        ).items():
            metadata: tuple[
                OperationEnum, FieldCriteria
            ] | None = self.__get_filterable_field_metadata(key)
            if metadata is None:
                continue
            assert isinstance(metadata, (tuple, list)) and len(metadata) == 2, (
                f"Field '{key}' must have metadata defined as a tuple of "
                f"(operation, metadata)."
            )
            if (conditions := self.__build_operation(*metadata, value)) is not None:
                filters.append(conditions)
        return filters if filters else None



from typing import Annotated, Any, Dict, Optional

from pydantic.fields import Field

from fastapi_advanced_filters.data_classes import AdvancedQSearch, QSearch


def generate_annotations_for_qsearch(
    filter_config_cls: type,
) -> Optional[Dict[str, Any]]:
    if hasattr(filter_config_cls, "q_search") and isinstance(
        filter_config_cls.q_search, (QSearch, AdvancedQSearch)
    ):
        q_search: QSearch | AdvancedQSearch = filter_config_cls.q_search
        return {
            "q_search": Annotated[
                str | None,
                Field(
                    default=None,
                    title="Search by query string",
                ),
                q_search,
            ]
        }
    return None



from fastapi_advanced_filters.data_classes import Pagination


class PaginationMixin:
    def build_pagination(self) -> Pagination | None:
        if hasattr(self, "limit") and hasattr(self, "offset"):
            return Pagination(limit=self.limit, offset=self.offset)
        if hasattr(self, "page") and hasattr(self, "page_size"):
            return Pagination(
                limit=self.page_size, offset=(self.page - 1) * self.page_size
            )
        return None



"""

Filter metaclass for creating filter configurations dynamically.
This module provides a metaclass that processes a nested `FilterConfig`
class within a Pydantic model to generate filter fields with
appropriate types, aliases, and metadata for filtering operations.
It supports features like field operations, sorting, selection,
full-text search, and pagination.

"""
from fastapi_advanced_filters.filter_metaclass.metaclass import FilterMetaClass



from dataclasses import dataclass
from typing import Any

from .pagination import Pagination


@dataclass(frozen=True)
class FilterResult:
    selected_columns: list[Any] | None = None
    sorting: list[Any] | None = None
    filters: list[Any] | None = None
    pagination: Pagination | None = None
    q_search: Any | None = None



from typing import Any, Callable

from pydantic import BaseModel, ConfigDict

from fastapi_advanced_filters.data_classes import FilterResult
from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum, OrderEnum
from fastapi_advanced_filters.filter_metaclass import FilterMetaClass
from fastapi_advanced_filters.filters.mixins import (
    FilterMixin,
    PaginationMixin,
    QSearchMixin,
    SelectMixin,
    SortingMixin,
)
from fastapi_advanced_filters.operation_mapping import (
    SQLALCHEMY_LOGICAL_OP_MAPPING as LOGICAL_OP_MAPPING,
)
from fastapi_advanced_filters.operation_mapping import (
    SQLALCHEMY_OP_MAPPING as OP_MAPPING,
)
from fastapi_advanced_filters.operation_mapping import (
    SQLALCHEMY_SORTING_MAPPING as SORTING_MAPPING,
)


class BaseFilter(
    BaseModel,
    FilterMixin,
    SortingMixin,
    SelectMixin,
    QSearchMixin,
    PaginationMixin,
    metaclass=FilterMetaClass,
):
    # Explicit types to satisfy mypy against the mixins' expectations
    __op_mapping__: dict[OperationEnum, Callable[..., Any]] = OP_MAPPING
    __sorting_mapping__: dict[OrderEnum, Callable[[Any], Any]] = SORTING_MAPPING
    __logical_op_mapping__: dict[
        LogicalOperator, Callable[..., Any]
    ] = LOGICAL_OP_MAPPING

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=False,
    )

    def get_filter_model(self) -> FilterResult:
        return FilterResult(
            filters=self.build_filters(),
            sorting=self.build_sorting(),
            pagination=self.build_pagination(),
            selected_columns=self.build_selectable_fields(),
            q_search=self.build_q_search(),
        )



from typing import Annotated, Any, Dict, Optional

from pydantic import Field, PlainValidator

from fastapi_advanced_filters.data_classes import SortBy
from fastapi_advanced_filters.utils import validate_sortable_schema


def generate_annotations_for_sortable_fields(
    filter_config_cls: SortBy | type,
) -> Optional[Dict[str, Any]]:
    if (
        hasattr(filter_config_cls, "sort_by")
        and isinstance(filter_config_cls.sort_by, SortBy)
        and (sort_by := filter_config_cls.sort_by)
    ):
        return {
            "sorting": Annotated[
                str | None,
                Field(
                    default=None,
                    alias="sort_by",
                    title=(
                        "Order by specific fields ("
                        f"{', '.join(sort_by.get_names())})"
                    ),
                    description=(
                        "Comma separated fields to order by. Prefix with '-' for"
                        " descending order."
                    ),
                ),
                sort_by,
                PlainValidator(validate_sortable_schema(sort_by.get_names())),
            ]
        }
    return None



"""
Operation mappings for different ORMs.

Currently, only SQLAlchemy is supported.
"""

from fastapi_advanced_filters.operation_mapping.sqlalchemy_mapping import (
    LOGICAL_OP_MAPPING as SQLALCHEMY_LOGICAL_OP_MAPPING,
)
from fastapi_advanced_filters.operation_mapping.sqlalchemy_mapping import (
    OP_MAPPING as SQLALCHEMY_OP_MAPPING,
)
from fastapi_advanced_filters.operation_mapping.sqlalchemy_mapping import (
    SORTING_MAPPING as SQLALCHEMY_SORTING_MAPPING,
)

__all__ = [
    "SQLALCHEMY_OP_MAPPING",
    "SQLALCHEMY_SORTING_MAPPING",
    "SQLALCHEMY_LOGICAL_OP_MAPPING",
]



from fastapi_advanced_filters.filters.base import BaseFilter
from fastapi_advanced_filters.filters.mixins import (
    FilterMixin,
    PaginationMixin,
    QSearchMixin,
    SelectMixin,
    SortingMixin,
)

__all__ = [
    "BaseFilter",
    "FilterMixin",
    "SortingMixin",
    "SelectMixin",
    "QSearchMixin",
    "PaginationMixin",
]



from dataclasses import dataclass, field
from typing import Any, Callable

from annotated_types import BaseMetadata

from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum
from fastapi_advanced_filters.utils import to_camel_case


@dataclass(frozen=True)
class FieldCriteria(BaseMetadata):
    name: str
    field_type: type
    op: tuple[OperationEnum, ...]
    model_attr: Any = None
    model_attrs_with_logical_op: tuple[list[Any], LogicalOperator] | None = None
    required_op: tuple[OperationEnum, ...] | None = None
    prefix: str | None = None
    custom_filter_per_op: Callable | None = None
    alias_as_camelcase: bool = False
    snake_case_separator_between_prefix_and_op: str = "__"
    fields_kwargs: dict[str, Any] = field(default_factory=dict)

    def get_name(self) -> str:
        if self.prefix is None:
            return self.name
        return (
            f"{self.prefix}{self.snake_case_separator_between_prefix_and_op}{self.name}"
        )

    def get_field_name(self, op: OperationEnum) -> str:
        if self.prefix is None:
            return (
                f"{self.name}"
                f"{self.snake_case_separator_between_prefix_and_op}"
                f"{op}"
            )
        return (
            f"{self.prefix}"
            f"{self.snake_case_separator_between_prefix_and_op}"
            f"{self.name}"
            f"{self.snake_case_separator_between_prefix_and_op}"
            f"{op}"
        )

    def get_alias_name(self, op: OperationEnum) -> str:
        if self.alias_as_camelcase:
            return to_camel_case(self.get_field_name(op))
        return self.get_field_name(op)



from dataclasses import dataclass
from typing import Any

from annotated_types import BaseMetadata

from fastapi_advanced_filters.enums import LogicalOperator, OperationEnum


@dataclass(frozen=True)
class QSearch(BaseMetadata):
    model_attrs: list[Any]
    logical_op: LogicalOperator = LogicalOperator.OR
    op: OperationEnum = OperationEnum.ILIKE
